// |    $Id: segd2seg.c,v 1.5 2024/03/14 01:22:22 pm Exp $      |
/* |------------------------------------------------------------|
   | Basic Seismic Utilities (BSU)                              |
   | LANGUAGE: C                                                |
   | AUTHOR: P. MICHAELS    http://173.255.241.228              | 
   | FUNCTION NAME: segd2seg.c                                  |
   | DIRECTORY: bsu-3.0.3/src/C/bsegy                           |
   |------------------------------------------------------------|
   | purpose: read limited version of SEG-D files               |
   |          and convert to BSEGY format.                      |
   |------------------------------------------------------------|
   |  NOTICE:                                                   |
   |  Copyright (C) 2021 Paul Michaels                          |
   |  <paulmichaels@boisestate.edu>                             |
   |  This program is free software; you can                    |
   |  redistribute it and/or modify it under the terms          |
   |  of the GNU General Public License as published            |
   |  by the Free Software Foundation; either version           |
   |  3 of the License, or (at your option) any later           |
   |  version.  This program is distributed in the              |
   |  hope that it will be useful, but                          |
   |  WITHOUT ANY WARRANTY; without even the implied            |
   |  warranty of MERCHANTABILITY or FITNESS FOR A              |
   |  PARTICULAR PURPOSE.  See the GNU General Public           |
   |  License for more details.                                 |
   |  You should have received a copy of the GNU                |
   |  General Public License along with this program;           |
   |  if not, write to the Free Software Foundation,            |
   |  Inc., 675 Mass Ave, Cambridge, MA 02139, USA.             |
   |------------------------------------------------------------|
   |INLINE FUNCTIONS:                                           |
   |  getparm    get parameters from command line or interactive|
   |  outlst     echo check of input parameters to listing file |
   |                                                            |
   |------------------------------------------------------------|
   |sublibC4 functions:                                         |
   |  c_bsegin   read seismic trace                             |
   |  c_bsegout  write a seismic trace                          |
   |  in_chk     scan input file for samples, number traces etc.|
   |  bargrid    write progress bar grid to screen              |
   |  exbar      write progress bar to screen                   |
   |  cr_labl    write copyright label to screen                |
   |  fcr_labl   write copyright label to file                  |
   |  names      create an output file name from input file name|
   |------------------------------------------------------------|
   |WORK IN PROGRESS !! LIMITED SEG-D at this point             |
   | SEG-D format is assumed to be revision 1 in this           |
   | code.  Further, there are many possible situations with    |
   | a great many possible channel scans and data format.       |
   | This code is for a little endian machine and assumes       |
   | SEG-D headers are a mixture of packed BCD and integers.    |
   | NOTE: Only one data format is supported at this point.     |
   | One would edit the data trace reading function for other   |
   | data formats than 8058 (32 bit IEEE demultiplexed float    |
   | using IEEE 754-1985 floats, big endian)                    |
   |                                                            |
   | Tested on two instruments: Sercel and Geode                |
   | Geode work-a-rounds assume only General Block Headers 1, 2 |
   | (no General Block Header 3 with location data)             |
   | Also, Geode case assumes two Scan type headers, first      |
   | is Seismic Data (Sercel used this for AUX) and the second  |
   | is unused (that is the Seismic Data on Sercel).            |
   | NOTE: Geometrics Knowledge Base appears inaccurate for     |
   | Geode (maybe Ok for StratVisor?).                          |
   | https://http://support.geometrics.com/kb/questions.php?questionid=33
   | Geode-NZ-File-Formats.pdf                                  |
   | P.M. Wed Mar  6 04:20:13 PM MST 2024                       |
   |                                                            |
   |Reference: Geophysics vol. 59, No. 4 (April 1994),          |
   |pages 668-684, Richard Allen et al.                         |
   |Digital field tape format standards SEG-D Revision 1        |
   |------------------------------------------------------------|
*/

// scan binary coded decimal (Packed BCD)
      #include <stdio.h>
      #include <stdlib.h>
      #include <float.h>
      #include <math.h>
      #include <string.h>
      #include <time.h>
      #include "../include/c_bsegy.h"

//...function prototypes
      #include "../include/sub4.h"

	int  getparm(int argc,char *argv[],
         char infile[],
         char ofile[], char ofile1[], char lstfile[],char pid[]);

	void outlst(char infile[],
        char pid[], char ofile[], char lstfile[], FILE *h3);

	void  assignhd(struct header *hd, int lineno, int vp, int sp,
	 int npts, float SF, int isamin, int filenumber,
	 short int highcut, short int lowcut, short int mcode,
	 short int tmin, short int thour, short int tsec,
	 short int tyear, short int tday, short int stack);


	float makefloat(unsigned int x);
	
	char BCD2ASCII(unsigned char original);

	short int GetIval(FILE *h1, long byteloc, unsigned char nout);

	void swap4(void *x4);
	
	void swap2(void *x2);

	void readheadexj(int j, int loc,  FILE *h1,
	int *lineno, int *vp );

	int  readtrace(int j, int locfirsthead,
	 int locfirstdata, int bytestrace,
	 unsigned short npts, float SF, float s1[],int *lineno,
	 int *vp, FILE *h1);

//---------------------end prototypes----------------------------------


int	main(int argc, char *argv[]) {
	FILE	*h1, *h2, *h3;
      int               jt,
                        jrec,
                        npts=0,
                        ntraces,
                        ibar,
                        invbar;
      float             fsamin,
                        bar,
                        *s1,
                        *s2;
      char              pid[5]="segd",   //process id name (4char)
                        ofile[13],  //main seismic data file
                        ofile1[13], //auxillary output file from scan 1
                        lstfile[13],
                        infile[14];
      struct header     hd;

//...specific to this application
	short int	filenumber;
	short int	formatcode;
	short int	dayofyear; 
	short int	year;
	unsigned char	ngbh; //number of additional general block headers
	short int	thour, tmin, tsec;
	short int	tyear, tday;
	short int	mcode;
	short int	nscantypes;
	short int	nchannelsets;
	short int	exthdlength;
	short int	externalhdlen;
	unsigned int	expfn; //expanded file number, binary 24 bits
	unsigned char 	revno; //segd revision number revno.pointrevno
	unsigned char 	pointrevno; 
	unsigned int	tmaxms; //record length in milliseconds
	unsigned char	gbhno; //general block header number 
	int		sourceln; //source line number
	int		spn; //location of source
	unsigned char	spidx; //source point index, increases with each move
	int		dt; //sample interval in microseconds
	unsigned int	channel1;
	unsigned short	startime;
	unsigned short	endtime;
	unsigned short	ntraces1;
	unsigned char	chtype1;
	unsigned char	vstack1;
	unsigned int	channel2;
	unsigned short	ntraces2;
	unsigned char	chtype2;
	unsigned char	vstack2;
	unsigned int	scanhed1indx;
	unsigned int 	extheadindx;
	unsigned int	firsttracehd;
	unsigned int	firsttracehd2;
	unsigned char	ntrhdext;
	unsigned int	firstdata;
	unsigned int	firstdata2;
	unsigned int	j, i; //index
	float		fdt;
	float		sample;
	int		datalength;
	int		bytestrace;
	long int	tracej;
	long int	headj;
	int		lengthtrhead;
	int		lineno, vp;
	unsigned char		MP1; //descale factor exponent, scan 1
	unsigned char		MP2; //descale factor exponent, scan 2
	int		isamin;
	char		gainmode;
	short int	highcut, highcut2;
	short int	hcslope, hcslope2;
	short int	lowcut, lowcut2;
	short int	lcslope, lcslope2;
	short int	goffset= 0;
//...Geode SEGD additions
	long int	BytesInFile;  // Use: determine number of samples/trace Geode
	int		BytesGH=64;      // bytes used by General Header Blocks
	int		BytesSH=64;      // bytes used by Scan Type Header Blocks
	int		BytesTR;      // bytes used by each trace (Demux Trace Hdr + data)
	int		BytesDMXH=20; // bytes used by Demux Trace Header, each trace 
	int		BytesPerSample=4; //WARNING: Assumes 32 bit demuxed data
	short int	stack;

//...get parameters
      if(getparm(argc,argv,
         infile,ofile,ofile1,lstfile,pid)!=0)
	{ goto quick_exit; }
     
//...open output listing file
        if( (h3=fopen(lstfile,"w"))==NULL )
        {
          fprintf(stderr,"\n ABORT !! Error opening file %s",lstfile);
          abort();
        }

//...echo check to listing file
      outlst(infile,pid,ofile,lstfile,h3);

//...open input file
        if( (h1=fopen(infile,"rb"))==NULL )
        {
          printf("\n ABORT !! Error opening file %s",infile);
	  abort();
        }
 
//...Read Header Values SEG-D File
	/* General Header 1 */
	fprintf(h3,"GENERAL HEADER #1 \n");
	filenumber = GetIval(h1,0,4);
	fprintf(h3,"File Number: %d \n",filenumber);

	formatcode = GetIval(h1,2,4);
	fprintf(h3,"Format Code: %d \n",formatcode);	
	if (formatcode == 8058 ) { fprintf(h3," 32 bit IEEE demultiplexed \n");}
	else {
	fprintf(stderr,"ABORT: Format code =%d Not Supported yet \n",formatcode);
	abort();}

	year = GetIval(h1,10,2);
	fprintf(h3,"Year 20%d \n",year);
	tyear = year;

	ngbh = GetIval(h1,11,2); // //Unpack BCD needs at least 2 digits
	ngbh = ngbh/10;
	fprintf(h3,"Number of General Block Headers = %d + 1 \n",ngbh);

	dayofyear = GetIval(h1,11,4);
	dayofyear = dayofyear - ngbh*1000;
	fprintf(h3,"Julian Day of Year %d \n",dayofyear);
	tday = dayofyear;

	thour = GetIval(h1,13,2);
	tmin  = GetIval(h1,14,2);
	tsec  = GetIval(h1,15,2);
	fprintf(h3,"Time: %2.2d:%2.2d:%2.2d \n",thour,tmin,tsec);

	mcode = GetIval(h1,16,2);
	fprintf(h3,"Manufacturer Code: %d \n",mcode);
	if (mcode == 13 ) fprintf(h3,"Sercel Instruments ");
	if (mcode == 0  ) fprintf(h3,"Geometrics Geode ");

	nscantypes = GetIval(h1,27,2);
	fprintf(h3,"Number of Scan Types per record = %d \n",nscantypes);

	nchannelsets = GetIval(h1,28,2);
	fprintf(h3,"Number of Channel Sets per scan type = %d \n",nchannelsets);

	exthdlength = GetIval(h1,30,2);
	fprintf(h3,"Extended Header Length = %d x 32 bytes = %d \n",
	exthdlength,exthdlength*32);

	externalhdlen = GetIval(h1,31,2);
	fprintf(h3,"External Header Length = %d x 32 bytes = %d \n",
	externalhdlen,externalhdlen*32);
	fprintf(h3,"END General Block Header 1 \n");
	fprintf(h3,"--------------------------------------------------\n");

	/* General Header 2  */
	fprintf(h3,"GENERAL HEADER #2 \n ");
	fseek(h1,32,SEEK_SET);
	fread(&expfn,sizeof(int), 1, h1);
	swap4(&expfn); //convert from big endian to little endian
	expfn=expfn >> 8; // shift right a byte (3*8=24 bit number)
	if (mcode == 0) expfn = filenumber;
	fprintf(h3,"Expanded File Number = %d \n",expfn);

	fseek(h1,42,SEEK_SET);
	fread(&revno,sizeof(char), 1, h1); 
	fread(&pointrevno,sizeof(char), 1, h1);
	fprintf(h3,"SEG-D Revision Number = %d.%2.2d \n",revno,pointrevno);

	fseek(h1,47,SEEK_SET);
	fread(&tmaxms,sizeof(int), 1, h1); 
	swap4(&tmaxms); //convert from big endian to little endian
	tmaxms=tmaxms >> 16 ; // shift right a byte (3*8=24 bit number)
	fprintf(h3,"Record Length = %d milliseconds \n",tmaxms);

	fseek(h1,50,SEEK_SET);
	fread(&gbhno,sizeof(char), 1, h1);
	fprintf(h3,"END General Block Header Number = %d \n",gbhno);
	fprintf(h3,"--------------------------------------------------\n");

	// Code only designed for up to 2 Channel Sets. Following only if 2.
	/* Additional General Header N */
	if (ngbh == 2) {
	fprintf(h3,"GENERAL HEADER #3 \n ");
	fseek(h1,69,SEEK_SET);
	fread(&sourceln,sizeof(int),1, h1);
	fprintf(h3,"Source Line Number = %d \n",sourceln);

	fseek(h1,71,SEEK_SET);
	fread(&spn,sizeof(int),1,h1);
	swap4(&spn);
	fprintf(h3,"Source Point Number =%d \n",spn);

	fseek(h1,77,SEEK_SET);
	fread(&spidx,sizeof(char),1, h1);
	fprintf(h3,"Source Point Index = %d \n",spidx);

	fseek(h1,82,SEEK_SET);
	fread(&gbhno,sizeof(char), 1, h1);
	fprintf(h3,"END General Block Header Number = %d \n",gbhno);
	fprintf(h3,"--------------------------------------------------\n");
	} //endif ngbh

/*	If only 1 channel scan set, ngbh =1, and above additional general
	header block will be absent.  This requires a negative 32 byte offset
	for what follows.
*/
	if (ngbh < 2 ) goffset = -32;

	/* Scan Block Header #1  */
	fprintf(h3,"SCAN BLOCK HEADER #1\n ");
	channel1 = GetIval(h1,97+goffset,2);	
	fprintf(h3,"Channel Set = %d \n",channel1);

	fseek(h1,98+goffset,SEEK_SET);
	fread(&startime,sizeof(short int), 1, h1);
	swap2(&startime);
	fprintf(h3,"Start Time =%u \n",startime);

	fseek(h1,100+goffset,SEEK_SET);
	fread(&endtime,sizeof(short int), 1, h1);
	swap2(&endtime);
	endtime=endtime*2; //2 msec steps
	fprintf(h3,"End Time =%d msec. \n",endtime);

	/* Scale Factor multiplier, scales to mv scan1*/
	float SF;
	/* Scale Factor multiplier, scales to mv scan2*/
	float SF2;

	// Sercel Instrument
	if ( mcode ==13 )
	{
	fseek(h1,103+goffset,SEEK_SET);
	fread(&MP1,sizeof(char),1,h1);
	fprintf(h3,"Descale Factor MP1=0X%0X  \n",MP1);
	char S= (MP1 >> 7) & 1; //fprintf(h3,"Signbit = %0x \n",S);
	char E= (MP1 >> 2) & 0x1f; // fprintf(h3,"Exponent Int = %d \n",E);
	char F= ((MP1 << 4) >>4) &0x3  ; //fprintf(h3,"Fraction = 0.%d \n",F);
	float SFexp = powf(-1.0,(float) S)*((float) E + (float) F/10.);
	SF = powf(2.,SFexp);
	fprintf(h3,"Scale Factor: Exponent = %f  SF =  %g \n",SFexp, SF);
	} // endif Sercel Instruments

	// Geode Instrument
	if (mcode == 0 ) 
	{
	int ISF;
	fseek(h1,60,SEEK_SET);
	fread(&ISF,sizeof(int),1,h1);
	SF = (float) ISF;
	fprintf(h3,"Scale Factor:  SF =  %g \n",SF);
	} // endif Geode Instruments


//...Get the number of traces
	ntraces1 = GetIval(h1,104+goffset,4);
	fprintf(h3,"Number of traces this scan = %d \n",ntraces1);

	fseek(h1,106+goffset,SEEK_SET);
	fread(&chtype1,sizeof(char), 1, h1);
	chtype1 =chtype1 >> 4;
	if (chtype1 == 9 ) fprintf(h3,"Channel Type is Signature/Filtered\n");
	if (chtype1 == 1 ) fprintf(h3,"Channel Type is Seismic\n");

	gainmode=GetIval(h1,107+goffset,2);
	if (gainmode == 3 ) fprintf(h3,"Gain Mode = %d Fixed Gain \n",gainmode);
	if (gainmode == 4 ) fprintf(h3,"Gain Mode = %d Programmed Gain \n",gainmode);
	if (gainmode == 8 ) fprintf(h3,"Gain Mode = %d Binary Gain Control \n",gainmode);

	highcut = GetIval(h1,108+goffset,4);
	fprintf(h3,"High Cut Filter (anti-alias) = %d Hz \n",highcut);
	hcslope = GetIval(h1,110+goffset,4);
	fprintf(h3,"High Cut Filter slope = %d dB/oct \n",hcslope);
	if (hcslope == 0 ) fprintf(h3,"---> High Cut filter Out \n");
	lowcut = GetIval(h1,112+goffset,4);
	fprintf(h3,"Low Cut Filter = %d Hz \n",lowcut);
	lcslope = GetIval(h1,114+goffset,4);
	fprintf(h3,"Low Cut Filter slope = %d dB/oct  \n",lcslope);
	if (lcslope == 0 ) fprintf(h3,"---> Low Cut filter Out \n");

	fseek(h1,125+goffset,SEEK_SET);
	fread(&vstack1,sizeof(char), 1, h1);
	fprintf(h3,"Vertical Stack = %d \n",vstack1);
	fprintf(h3,"END Scan Block Header 1\n ");
	fprintf(h3,"--------------------------------------------------\n");

	if (ngbh == 2  )
	{
	fprintf(h3,"SCAN BLOCK HEADER #2\n ");
	channel2 = GetIval(h1,129+goffset,2);	
	fprintf(h3,"Channel Set = %d \n",channel2);

	fseek(h1,130+goffset,SEEK_SET);
	fread(&startime,sizeof(short int), 1, h1);
	swap2(&startime);
	fprintf(h3,"Start Time =%u \n",startime);

	fseek(h1,132+goffset,SEEK_SET);
	fread(&endtime,sizeof(short int), 1, h1);
	swap2(&endtime);
	endtime=endtime*2; //2 msec steps
	fprintf(h3,"End Time =%d msec. \n",endtime);

	if (mcode == 13 )
	{
	/* Scale Factor multiplier, scales to mv scan 2*/
	fseek(h1,135+goffset,SEEK_SET);
	fread(&MP2,sizeof(char),1,h1);
	fprintf(h3,"Descale Factor MP2=0X%0X  \n",MP2);
	char S2= (MP2 >> 7) & 1; //fprintf(h3,"Signbit = %0x \n",S2);
	char E2= (MP2 >> 2) & 0x1f; // fprintf(h3,"Exponent Int = %d \n",E2);
	char F2= ((MP2 << 4) >>4) &0x3  ; //fprintf(h3,"Fraction = 0.%d \n",F2);
	float SF2exp = powf(-1.0,(float) S2)*((float) E2 + (float) F2/10.);
	SF2 = powf(2.,SF2exp);
	fprintf(h3,"Scale Factor: Exponent = %f  SF2 =  %g \n",SF2exp, SF2);
	}

	// Geode Instrument
	if (mcode == 0 ) 
	{
	SF2=SF;
	fprintf(h3,"Scale Factor:  SF2 =  %g \n",SF2);
	} // endif Geode Instruments

	ntraces2 = GetIval(h1,136,4);
	fprintf(h3,"Number of traces this scan = %d \n",ntraces2);

	fseek(h1,138+goffset,SEEK_SET);
	fread(&chtype2,sizeof(char), 1, h1);
	chtype2 =chtype2 >> 4;
	if (chtype2 == 9 ) fprintf(h3,"Channel Type is Signature/Filtered\n");
	if (chtype2 == 1 ) fprintf(h3,"Channel Type is Seismic\n ");

	gainmode=GetIval(h1,139+goffset,2);
	if (gainmode == 3 ) fprintf(h3,"Gain Mode = %d Fixed Gain \n",gainmode);
	if (gainmode == 4 ) fprintf(h3,"Gain Mode = %d Programmed Gain \n",gainmode);
	if (gainmode == 8 ) fprintf(h3,"Gain Mode = %d Binary Gain Control \n",gainmode);

	highcut2 = GetIval(h1,140+goffset,4);
	fprintf(h3,"High Cut Filter (anti-alias) = %d Hz \n",highcut2);
	hcslope2 = GetIval(h1,142+goffset,4);
	fprintf(h3,"High Cut Filter slope = %d dB/oct \n",hcslope2);
	if (hcslope2 == 0 ) fprintf(h3,"---> High Cut filter Out \n");
	lowcut2 = GetIval(h1,144+goffset,4);
	fprintf(h3,"Low Cut Filter = %d Hz \n",lowcut2);
	lcslope2 = GetIval(h1,146+goffset,4);
	fprintf(h3,"Low Cut Filter slope = %d dB/oct \n",lcslope2);
	if (lcslope2 == 0 ) fprintf(h3,"---> Low Cut filter Out \n");

	fseek(h1,157+goffset,SEEK_SET);
	fread(&vstack2,sizeof(char), 1, h1);
	fprintf(h3,"Vertical Stack = %d \n",vstack2);
	fprintf(h3,"END Scan Block Header 2 \n");
	fprintf(h3,"--------------------------------------------------\n");
	} //endif ngbh ==2

	/* Start fseek Indexes  */
	fprintf(h3," Number of samples and sample interval\n");
	if (ngbh == 2 ) {
	scanhed1indx = 3*32;}
	else {
	scanhed1indx = 2*32;
	} //endif

	extheadindx = scanhed1indx + nchannelsets*32;

	/* First Scan */
	firsttracehd = extheadindx + exthdlength*32 + externalhdlen*32;
	/* Move to the first trace header start, Read number of 32byte extensions */
	fseek(h1,firsttracehd+9,SEEK_SET);
        fread(&ntrhdext,sizeof(char),1, h1); 

	/* Compute trace header plus extensions, add to first trace header location
	and this is the start of the first data trace */
	lengthtrhead = 20 + ntrhdext*32;
	firstdata = firsttracehd + lengthtrhead;

	/* Move to the number of samples value in the trace header
	read the number of samples and byte swap */
	fseek(h1,firsttracehd+28,SEEK_SET);
	fread(&npts,sizeof(short int),1,h1);
	swap2(&npts);

	// npts computation for Geode Case
	if (mcode == 0 ) 
	{
	fseek(h1,0,SEEK_END);
	BytesInFile = ftell(h1);
	fprintf(h3,"File Size in bytes = %ld \n",BytesInFile);
	BytesTR = (BytesInFile - BytesGH - BytesSH)/ntraces1;
	npts = (BytesTR -BytesDMXH)/BytesPerSample;
	} //endif mcode=0 npts calculation
	fprintf(h3,"Number of samples =%d \n",npts);

	/* Compute sample interval from Scan Block Headers */
	dt = ((endtime - startime)*1000)/(npts-1);
	fprintf(h3,"Sample Interval = %d micro seconds \n",dt);
	fdt = ((float) dt)/1000000.; //sample interval in seconds 
	isamin =(int) floor(fdt*1.0E+6); //microseconds
	datalength = npts * 4;
	bytestrace = lengthtrhead + datalength;
	fprintf(h3,"Number of bytes per trace = %d \n",bytestrace);

	/* Second Scan */
	firsttracehd2= firsttracehd + ntraces1*bytestrace;
	firstdata2 = firstdata + ntraces1*bytestrace;	
	fprintf(h3,"--------------------------------------------------\n");

	if (mcode == 13 )
	{
	fprintf(h3," Start Indexes in file (bytes)\n");
	fprintf(h3,"Scan 1: First Trace Header Index = 0X%0X \n",firsttracehd);
	fprintf(h3,"Scan 1: First Data Sample Index = 0X%0X \n",firstdata);
	fprintf(h3,"Scan 2: First Trace Header Index = 0X%0X \n",firsttracehd2);
	fprintf(h3,"Scan 2: First Data Sample Index = 0X%0X \n",firstdata2);
	fprintf(h3,"--------------------------------------------------\n");
	} //endif mcode 13 Sercel

	if (mcode == 0 )
	{
	fprintf(h3," Start Indexes in file (bytes)\n");
	fprintf(h3,"Scan 1: First Trace Header Index = 0X%0X \n",firsttracehd);
	fprintf(h3,"Scan 1: First Data Sample Index = 0X%0X \n",firstdata);
	fprintf(h3,"Scan 2 Not Used \n");
	fprintf(h3,"--------------------------------------------------\n");
	} //endif mcode 0 Geode

        s1=calloc(npts,sizeof(float));
          if(s1==NULL)
         {
          fprintf(stderr,"ABORT:s1 can't allocate enough trace memory");
          fprintf(stderr," npts = %d",npts);
          abort();
         }

//=======================================================================
	/* Read Scan 1 DATA and set HEADER, OUTPUT to BSEGY */
	// assign output file names using the assumption that the first
	// scan set is the auxillary traces (IF Sercel Instruments).
	//  No way to really know.  For Geode scan1 = seismic data, no aux.
	if (ngbh == 2) {
	//...open auxillary output file
        if( (h2=fopen(ofile1,"wb"))==NULL )
	        {
        	  fprintf(stderr,"\n ABORT !! Error opening file %s",ofile1);
          	abort();
        	}
	puts("Processing Auxillary Traces");
	fprintf(h3,"AUXILLARY TRACES---------------------------\n");
	}
	else {
	//...open shot record 
        	if( (h2=fopen(ofile,"wb"))==NULL )
        	{
          	fprintf(stderr,"\n ABORT !! Error opening file %s",ofile);
          	abort();
        	}
	puts("Processing Seismic Traces");
	fprintf(h3,"SEISMIC TRACES---------------------------\n");
	} //endif ngbh

	for (j=0 ; j<ntraces1; j++) {
//	    trace numbers start at 0
	tracej = firstdata + j*bytestrace;
	headj  = firsttracehd + j*bytestrace; 
	if (readtrace(j,headj,tracej,bytestrace,npts,SF,s1,&lineno,&vp,h1)!= 0)
	fprintf(h3,"j=%d lineno=%d vp=%d \n",j,lineno,vp);
	else {
	fprintf(stderr,"Error reading input file, trace %d \n",j);
	abort();}

	/*     set headers   */
	//default for when Geode has no location info
	if (mcode == 0 ) {
	vp=j;
	spn=0;
	lineno=1;
	} // when Geode has no General Trailer or Trace Header Extension, location info

	stack=vstack1;
	assignhd(&hd,lineno,vp,spn,npts,SF,isamin,filenumber,
	highcut,lowcut,mcode,tmin,thour,tsec,tyear,tday,stack);

	//...output trace
        if(c_bsegout(j,s1,npts,&hd,h2)!=0)
       {
        puts("ABORT--c_bsegout");
        abort();
       }

//...display progress bar
        exbar(ntraces1,bar,ibar,invbar,j+1);
	} //next j
	fprintf(stdout,"\n %d traces processed \n",ntraces1);
	fclose(h2);
	ntraces=ntraces1;  //if only 1 scan, if 2 scans, updated below

//=======================================================================
	if (mcode == 13)
	{
	/* Read Scan 2 SATA and set HEADER, OUTPUT to BSEGY */
	// assign output file names using the assumption that the first
	// scan set is the auxillary traces.  No way to really know.
	if (ngbh == 2) {
	//...open output file
        if( (h2=fopen(ofile,"wb"))==NULL )
        {
          fprintf(stderr,"\n ABORT !! Error opening file %s",ofile);
          abort();
        }

	ntraces=ntraces1+ntraces2; //total number of traces	
	fprintf(stdout,"\nProcessing Seismic Traces\n");
	fprintf(h3,"SEISMIC TRACES---------------------------\n");
	for (j=0 ; j<ntraces2; j++) {
//	j=ntraces2 -1 ; //trace numbers start at 0
	tracej = firstdata2 + j*bytestrace;
	headj  = firsttracehd2 + j*bytestrace; 
	if (readtrace(j,headj,tracej,bytestrace,npts,SF,s1,&lineno,&vp,h1)!= 0)
	fprintf(h3,"j=%d lineno=%d vp=%d \n",j,lineno,vp);
	else {
	fprintf(stderr,"Error reading input file, trace %d \n",j);
	abort();}

	/*     set headers   */
	stack=vstack2;
	assignhd(&hd,lineno,vp,spn,npts,SF,isamin,filenumber,
	highcut2,lowcut2,mcode,tmin,thour,tsec,tyear,tday,stack);

	//...output trace
        if(c_bsegout(j,s1,npts,&hd,h2)!=0)
       {
        puts("ABORT--c_bsegout");
        abort();
       }

//...display progress bar
        exbar(ntraces2,bar,ibar,invbar,j+1);
	} //next j
	fprintf(stdout,"\n %d traces processed \n",ntraces2);
	fclose(h2);
	} //endif ngbh==2

	} //endif mcode==13
	fclose(h1);
	fclose(h3);
      printf("\n   Listing ====>%s",lstfile);
      printf("\n   Output  ====>%s",ofile);
      if(ngbh==2) printf("\n   Output Aux Traces ====>%s",ofile1);
      printf("\n   Total: Processed %d traces\n",ntraces);

	return(0);
quick_exit:
	return(1);

} //end main

/*            Read a trace function                      */
	int	readtrace(int j, int locfirsthead,
		int locfirstdata, int bytestrace,
		unsigned short npts, float SF, float s1[],int *lineno,
		int *vp, FILE *h1  )
	{
	int		i;
	unsigned int	buff;
	size_t		status;  //if zero, then error or eof
	/* j = trace number in current scan
	   locfirsthead is absolute position of first byte, this scan trheader
	   locfirstdata is absolute position of first byte, data trace
	   bytestrace = length of trace header (bytes) + data (bytes)
	   npts = number of samples in a trace 
	   SF = scale factor 2^MP multiply the trace samples by to get mv
	   *h1 = file assumed open with data
	RETURN VALUES, lineno, vp, s1
	*/
	//read trace header for line number and receiver VP
	readheadexj(j,locfirsthead,h1, lineno, vp);
	fseek(h1,locfirstdata,SEEK_SET);
	// Read a trace 
	for (i=0; i<npts; i++) {
	status = fread(&buff,sizeof(int),1,h1);
	swap4(&buff);
	s1[i] = makefloat(buff)*SF;
	} //next i

	return (status);	
	}  //end readtrace
 
/*            Function makefloat                         */
	float makefloat(unsigned int x)
	{
	unsigned char	masksign = 0x1;
	unsigned short	maskexp  = 0xff;
	unsigned int	maskfrac = 0x007fffff;
	float A, B, C;
	unsigned char  s;
	unsigned short e;
	unsigned int   f;

//fprintf(stdout," x = %08x \n",x);
	
	s = masksign& (x >> 31);
	e = maskexp& (x >>23);
	f = maskfrac&x;
	A = powf(-1.0,(float) s); 
	B = powf(2.0,(float) (e-127)); 
	C = (1.+((float)f)/powf(2.0,23.)); 

//fprintf(stdout," s=%d \n e=%d \n f=%d \n",s,e,f); 
	return (A*B*C);
	} //end makefloat

/*            Function BCD2ASCII                          */
/* Converts Packed BCD 8bit to 16 bits (two 8 bits)  */
	char BCD2ASCII(unsigned char original)
	{
	unsigned char tens, ones;
 	tens=original/16;
 	ones=original&15;
 	return (tens*10+ones); 
	} //end BCD2ASCII

/*            Function GetIval                            */
/*Returns unpacked interger from stream location byteloc  
  nout = number of of output digits */
	short int GetIval(FILE *h1, long byteloc, unsigned char nout)
	{
	char		val;
	unsigned char	x[2];
	short int	ival;
	unsigned char	ctr;
	int		i;

	fseek(h1,byteloc,SEEK_SET);
	for (i=0; i<2; i++) {
	fscanf(h1,"%c",&ctr);
	val=BCD2ASCII(ctr);
	x[i]=val;
	} // next i
	switch (nout) {
	case 4:
	ival=100*x[0]+x[1];
	break;
	case 3:
	ival=(x[0]/10)*100 + x[1];
	break;
	case 2:
	ival=x[0];
	break;
	} //endswitch

	return (ival);
	} //end GetIval

/* 4 byte  swap  PM */
	void swap4(void *x4)
	{
	unsigned char *tmp=(unsigned char *) x4;
	unsigned char x;
	x=tmp[0];
	tmp[0]=tmp[3];
	tmp[3]=x;
	x=tmp[1];
	tmp[1]=tmp[2];
	tmp[2]=x;
	} //end swap4
/* 2 byte  swap  PM */
	void swap2(void *x2)
	{
	unsigned char *tmp=(unsigned char *) x2;
	unsigned char x;
	x=tmp[0];
	tmp[0]=tmp[1];
	tmp[1]=x;
	} //end swap2

	void readheadexj(int j, int loc, FILE *h1,
	int *lineno, int *vp )
	{	
	fseek(h1,loc + 20 , SEEK_SET);
	fread(lineno,3,1,h1);
	swap4(lineno);
	*lineno = *lineno >>8;
	fseek(h1,loc + 23, SEEK_SET);
	fread(vp,3,1,h1);
	swap4(vp);
	*vp = *vp >>8;
	return ;
	}

/* |------------------------------------------------------------|
   | Basic Seismic Utilities (BSU)                              | 
   | LANGUAGE: C                                                |
   | AUTHOR: P. MICHAELS <paulmichaels@boisestate.edu>          |
   | FUNCTION NAME: getparm.c                                   |
   | Copyright (C) 2021 P. Michaels  (See GNU License)          |
   |------------------------------------------------------------|
   |                                                            |
   | purpose: gets input parameters from command                |
   |          line or prompts user from consol.                 |
   |                                                            |
   |                                                            |
   |------------------------------------------------------------|
*/
      int  getparm(int argc,char *argv[],
         char infile[],
         char ofile[], char ofile1[], char lstfile[],char pid[])
     {
      char              sfx1[4]="seg",
                        sfx2[4]="lst";
      char		help[3]="-h";
      char              msg1[46],msg2[46];

//...test for command line help ?
      if(argc==2 && !strcmp(argv[1],help))
     {
//...screen header and copyright notice
      cr_labl(pid);

//...help pannel
      sprintf(msg1,"Convert SEG-D to BSEGY format                ");
      sprintf(msg2," Very Limited Functionality 8058 Data Format ");
      hlp_labl(msg1,msg2);
      puts(" ");

//...help display
      puts(" ");
      printf("  %s2seg  infile    \n",pid);
      puts(" ");
      puts("     infile  = input file name (4char minimum,");
      puts("      13 character maximum )             ");
      puts("     Example:     xxxx.segd            ");
      puts("     (use symbolic link with shorter name if a problem) ");
      puts(" ");
      return(1);
     }
      else
     {
//...get parameters------------------------------
//...pars off character parameters
      if(argc>1)
       strncpy(infile,argv[1],13);
       else
      {
       puts(" Enter INPUT file name ");
       scanf(" %13s",infile);
      }
	if (namecheck(12,infile) != 0 ) return (1);

//...name output files
//      names(infile,pid,ofile,sfx1);
//      names(infile,pid,lstfile,sfx2);

//...name output files
	char	stmp[5];
	char	cnul=0;
	short int 	i;
       for (i=0;i<4;i++) stmp[i]=infile[i];
       stmp[4]=cnul;

       sprintf(ofile,"%s.seg",stmp); 
       sprintf(ofile1,"aux%s.seg",stmp); //auxillary output scan 1
       sprintf(lstfile,"%s.lst",stmp); 

       return(0);
     }  //  endif bhelp test
      } // end getparm

/* |------------------------------------------------------------|
   | Basic Seismic Utilities (BSU)                              | 
   | LANGUAGE: C                                                |
   | AUTHOR: P. MICHAELS <paulmichaels@boisestate.edu>          |
   | FUNCTION NAME: outlst.c                                    |
   | Copyright (C) 2021 P. Michaels  (See GNU License)          |
   |------------------------------------------------------------|
   |                                                            |
   | purpose: prints copyright header, file names, and          |
   |          echo check of input parameters in newly           |
   |          opened output listing file.                       |
   |                                                            |
   |------------------------------------------------------------|
*/
      void outlst(char infile[],
        char pid[], char ofile[], char lstfile[], FILE *h3)
     {

//...print copyright label to outlst and screen

      fcr_labl(pid,h3);

//...echo check
      fputs("\n",h3);                      //blank line
      fprintf(h3,"\n     PROGRAM: %s2seg.c ",pid);
      fputs("\n",h3);                      //blank line
      fputs("\n     FILES:       ",h3);
      fprintf(h3,"\n     Input   File= %s",infile);
      fprintf(h3,"\n     Output  File= %s",ofile);
      fprintf(h3,"\n     Listing File= %s",lstfile);
      fputs("\n",h3);                      //blank line
     }  //  end outlst

/*          ASSIGN HEADER VALUES          */
	void  assignhd(struct header *hd, int lineno, int vp, int sp, int npts,
	float SF, int isamin, int filenumber, short int highcut, 
	short int lowcut, short int mcode,short int tmin, short int thour,
	short int tsec, short int tyear, short int tday, short int stack)
{
	//...assign default header values
	char	blnk=' ';
	char	buff[10];
	short int	SFdb;
	if ( mcode == 0 )
	{
	SFdb = (short int) floor((20*log10f(SF))) + 1;
	}
	else {
	SFdb = (short int) floor((-1)*(20*log10f(SF)));
	} //endif mcode

	(*hd).idist=0;	//range, shot to receiver
	(*hd).length=(unsigned short) npts; //number of samples in trace
	(*hd).isi=(unsigned short) isamin;
	(*hd).duse=1;
	(*hd).ofrn= filenumber;
	(*hd).smul1=1;	//multiplier for next values
	(*hd).irel=0;	//receiver elevation
	(*hd).ishe=0;	//shot elevation
	(*hd).ishd=0;	//shot depth
	(*hd).delr=0;	//datum elevation of receiver
	(*hd).dels=0;	//datum elevation of source
	(*hd).wds=0;	//water depth of source
	(*hd).wdr=0;	//water depth of receiver
	(*hd).ictime=(short int) 0; //delaytime
	(*hd).tmin=tmin; // time of shot
	(*hd).thour=thour;
	(*hd).tsec=tsec;
	(*hd).tyear=tyear;
	(*hd).tday=tday;
	(*hd).nvs=stack;

	(*hd).smul2=1;	//multiplier for next values
	(*hd).ishlo=sp;	//shot coordinate, sx
	(*hd).ishla=0;	//shot coordinate, sy
	if (mcode == 13)
	(*hd).irlo=vp ;	//receiver coordinate, rx
	else
	(*hd).irlo=0 ;	//receiver coordinate, rx
	(*hd).irla=0;	//receiver coordinate, ry

	(*hd).cunits=1;	//units in meters
	(*hd).wvel=0;	//weathering velocity
	(*hd).swvel=0;	//subweathering velocity
	(*hd).utimes=0;	//uphole time at shot
	(*hd).utimeg=0;	//uphole time at group
	(*hd).sstati=0;	//shot static (ms)
	(*hd).gstati=0;	//group static (ms)
	(*hd).tstati=0;	//total static (ms)
	(*hd).istime=0;	//lag time A
	(*hd).ibtime=0;	//lag time B
	(*hd).mtimes=0;	//mute start time
	(*hd).mtimee=0;	//mute end time
	(*hd).tic=1;	//trace id code
	(*hd).gaint=3;  //Gain (1=fixed 2=binary 3=float)
  	(*hd).gidb=SFdb;    //Instrument gain in db
	(*hd).fhc=highcut;  //high cut filter
	(*hd).flc=lowcut;  //low cut filter
	sprintf(buff,"%c%c%c%c",blnk,blnk,blnk,blnk);
	memcpy(&(*hd).deploy[0],&buff[0],4);	//deployment name
	sprintf(buff,"%4.4d",sp);
	memcpy(&(*hd).spname[0],&buff[0],4);	//shot name
	sprintf(buff,"%4.4d",vp);
	memcpy(&(*hd).rstnam[0],&buff[0],4);	//receiver site name
	sprintf(buff,"%4.4d",sp);
	memcpy(&(*hd).shotid[0],&buff[0],4);	//shot id
	sprintf(buff,"%4.4d",lineno);
	memcpy(&(*hd).lineid[0],&buff[0],4);	//line name
	sprintf(buff,"%c%c%c%c",blnk,blnk,blnk,blnk);
	memcpy(&(*hd).geoor[0],&buff[0],4);	//geophone orientation
	(*hd).mst=0;	//first break pick time
	(*hd).shtazi=0;	//source azimuth East=90
	(*hd).charge=0;
	(*hd).shtver=180;	//source polarization down
	(*hd).azimut=0;
	(*hd).geoazi=0;	//geophone azimuth East=90
	(*hd).geover=0;
}  //end assignhd 
