// |$Id: wav2txt.c,v 1.1.1.1 2022/12/27 22:03:00 pm Exp $          |
/* |------------------------------------------------------------|
   | Basic Seismic Utilities (BSU)                              |
   | LANGUAGE: C                                                |
   | AUTHOR: P. MICHAELS    http://173.255.241.228              | 
   | FUNCTION NAME: wav2txt.c                                   |
   | DIRECTORY: bsu-3.0.3/src/C/bsegy                           |
   |------------------------------------------------------------|
   | purpose: convert audio *.wav file to ASCII txt file        |
   |------------------------------------------------------------|
   |  NOTICE:                                                   |
   |  Copyright (C) 2021 Paul Michaels                          |
   |  <paulmichaels@boisestate.boisestate.edu>                  |
   |  This program is free software; you can                    |
   |  redistribute it and/or modify it under the terms          |
   |  of the GNU General Public License as published            |
   |  by the Free Software Foundation; either version           |
   |  3 of the License, or (at your option) any later           |
   |  version.  This program is distributed in the              |
   |  hope that it will be useful, but                          |
   |  WITHOUT ANY WARRANTY; without even the implied            |
   |  warranty of MERCHANTABILITY or FITNESS FOR A              |
   |  PARTICULAR PURPOSE.  See the GNU General Public           |
   |  License for more details.                                 |
   |  You should have received a copy of the GNU                |
   |  General Public License along with this program;           |
   |  if not, write to the Free Software Foundation,            |
   |  Inc., 675 Mass Ave, Cambridge, MA 02139, USA.             |
   |------------------------------------------------------------|
   |INLINE FUNCTIONS:                                           |
   |  getparm    get parameters from command line or interactive|
   |  outlst     echo check of input parameters to listing file |
   |                                                            |
   |------------------------------------------------------------|
   |sublibC4 functions:                                         |
   |  c_bsegin   read seismic trace                             |
   |  c_bsegout  write a seismic trace                          |
   |  in_chk     scan input file for samples, number traces etc.|
   |  bargrid    write progress bar grid to screen              |
   |  exbar      write progress bar to screen                   |
   |  cr_labl    write copyright label to screen                |
   |  fcr_labl   write copyright label to file                  |
   |  names      create an output file name from input file name|
   |                                                            |
   |------------------------------------------------------------|
*/
      #include <stdio.h>
      #include <stdlib.h>
      #include <float.h>
      #include <math.h>
      #include <string.h>
      #include <time.h>
	#include <fcntl.h>
	#include<sys/stat.h>
      #include "../include/c_bsegy.h"

//...parameter definitions
	struct wav_header
	{				//				byte
	char		ID[4];  	//should be RIFF		0-3
	int		Size;		// total bytes minus 8		4-7
	char    	typ[8];   	// "WAVEfmt "			8-15
	int		PCM;		// 16 for PCM format		16-19
	short int	PCM2;		// 1 for PCM format 		20-21
	short int	nchannels;	// number of channels		22-23
	int		fsampling;	// sampling rate (samples/sec)	24-27
	int		nbps;		// number bytes per second	28-31
	short int	nbpcap;		// number bytes per capture	32-33
	short int	nbpsamp;	// number of bits per sample	34-35
	char		worddata[4];	// word "data" is ASCII		36-39
	int		bytesdata;	// number of bytes in data	40-43
	};
	

//...function prototypes
      #include "../include/sub4.h"

      int  getparm(int argc,char *argv[],
        char infile[],
        char ofile[], char lstfile[], int *n1, int *nlast, char pid[]);

      void outlst(char infile[],
        char pid[], char ofile[], char lstfile[], FILE *h3);

	long finddata(int *nbytesdata, FILE *h1  );


      int  main(int argc, char *argv[])
     {

//...declare variables
      FILE              *h1, *h2, *h3;
      int               jt,
                        jrec,
                        npts,
                        ntraces,
                        ibar,
                        invbar;
      float             fsamin,
                        bar;
      char              pid[5]="wav2",   //process id name (4char)
                        ofile[80],
                        lstfile[80],
                        infile[80];
//      struct header     hd;

//...specific to this application
	struct 		wav_header whd;
	int		n1,		//first sample to extract
			nlast,		//last sample to extract
			nextract, 	//number of samples to extract
			nbytesdata;
	long		gostart;
	float		jtime;  	//sample time
	short int	s1[2];  	// sample=[ch1 ch2] pairs in sequence
	struct stat sb;

//...get parameters
      if(getparm(argc,argv,
         infile,ofile,lstfile,&n1,&nlast,pid)!=0)
      {
       goto quick_exit;
      }

//...open output listing file
        if( (h3=fopen(lstfile,"w"))==NULL )
        {
          fprintf(stderr,"\n ABORT !! Error opening file %s",lstfile);
          goto quick_exit;
        }

//...echo check to listing file
      outlst(infile,pid,ofile,lstfile,h3);

//...open input file
        if( (h1=fopen(infile,"rb"))==NULL )
        {
          fprintf(stderr,"\n ABORT !! Error opening file %s",infile);
          goto exit_3;
        }

//...open output file
        if( (h2=fopen(ofile,"wb"))==NULL )
        {
          fprintf(stderr,"\n ABORT !! Error opening file %s",ofile);
          goto exit_1;
        }

//...get file size in bytes
	if (stat(infile, &sb)  == -1 ) {
	perror("stat error");
	abort();
	}
//	printf("File Size, %lu  bytes \n",sb.st_size);

//... find "data" string, then read int nbytesdata that immediately
//    follows data string
	gostart = finddata(&nbytesdata,h1);

//read header of wav file
	rewind(h1);
	        if(fread(&whd,sizeof(char),44,h1)!=44)
            {
             fputs("ABORT--: Error reading input header ",stderr);
             return(1);
            }

//...print input file data to lstfile
	fprintf(h3,"\n File: %s \n",infile);
	fprintf(h3," ID = %4.4s \n",whd.ID);
	fprintf(h3," Size less 8 = %d  \n",whd.Size );
	fprintf(h3," Type = %8.8s \n",whd.typ );
	fprintf(h3," PMC = %d     \n",whd.PCM );
	fprintf(h3," PMC2 = %d     \n",whd.PCM2 );
	fprintf(h3," \n");
	fprintf(h3," Number of Channels:          nchannels=%d \n",whd.nchannels);
	fprintf(h3," Sampling Rate:               fsampling=%d \n",whd.fsampling);
	fprintf(h3," Number of bytes by capture:  nbpcap=%d \n",whd.nbpcap);
	fprintf(h3," Number of bits per sample:   nbpsamp=%d \n",whd.nbpsamp);
	fprintf(h3," worddata = %4.4s     \n",whd.worddata );

	if (whd.bytesdata != nbytesdata) whd.bytesdata = nbytesdata;
	fprintf(h3," Number of bytes in data:     bytesdata=%d \n",whd.bytesdata);
	npts=whd.bytesdata/whd.nbpcap;
	fprintf(h3," Number of samples in data:   npts=%d \n",npts);
	fprintf(h3," Check on Number of samples = %d \n",(sb.st_size-(gostart+4))/whd.nbpcap);

//...compute sample interval
	fsamin=1/ (float) whd.fsampling;
        fprintf(h3,"\n     Sample Interval = %14.7E sec. \n",fsamin);
        ntraces=whd.nchannels;

//...confirm test specific type of wav file
	if (whd.nbpcap/whd.nchannels !=2 || whd.nbpsamp != 16) {
	fprintf(stderr,"ABORT: code only written for 2 byte short samples \n");
	fprintf(stderr,"number of bytes = %d  number of bits = %d \n",
	whd.nbpcap,whd.nbpsamp);
	abort();
	}

//...echo check on range to extract
	float 	T1 = n1 * fsamin;
	float	T2 = nlast * fsamin;
	fprintf(h3,"     First sample number to extract = %d (%.10f sec) \n",n1,T1);
	fprintf(h3,"     Last  sample number to extract = %d (%.10f sec) \n",nlast,T2);
	fflush(h3);


//...get current position for first sample available
	gostart = gostart + 4;  //location of first sample, number 1
	gostart = gostart + (whd.nbpcap)*(n1-1);
	fseek(h1,gostart,SEEK_SET);  //position file pointer to first sample to extract
//	printf("debug: gostart = %ld \n",gostart);

//-----------------main loop---------------------------
//...display progress bar grid
	nextract = nlast-n1 + 1 ;
//	printf("debug: nextract = %d \n",nextract);

      bargrid(nextract/ntraces,&bar,&ibar,&invbar);

//...print column headers
	if (ntraces == 1)
	fprintf(h2,"# Sample \tTime \t\tChannel_1 \n");
	else
	fprintf(h2,"# Sample \tTime \t\tChannel_1 \tChannel_2 \n");


//...sample,trace loop

      for (jt=0;jt<nextract;jt++)
     {
//...read trace
	if(fread(s1,sizeof(short int),ntraces,h1) != ntraces)
	 {
	 fputs("ERROR reading data \n",stderr);
	  return(1);
	} //endif

//...output trace
	jtime = ((float)(n1+jt))*fsamin;
	if (ntraces == 1) {
	fprintf(h2,"  %10.10d \t%f \t%d \n",n1+jt,jtime,s1[0]);}
	else {
	fprintf(h2,"  %10.10d \t%f \t%d \t\t%d \n",n1+jt,jtime,s1[0],s1[1]);
	} //endif

//...display progress bar
      exbar(npts,bar,ibar,invbar,jt);

     } //  end trace loop-----------------------------

      printf("\n   Listing ====>%s",lstfile);
      printf("\n   Output  ====>%s",ofile);
      printf("\n   Processed %d traces \n",ntraces);
        return(0);

quit_it:
        puts(" ");
exit_s1:
        free(s1);
exit_2:
        fclose(h2);
exit_1:
        fclose(h1);
exit_3:
        fclose(h3);
quick_exit:
        return(1);
     } //  end of main
/* |------------------------------------------------------------|
   | Basic Seismic Utilities (BSU)                              | 
   | LANGUAGE: C                                                |
   | AUTHOR: P. MICHAELS <paulmichaels@boisestate.boisestate.edu>|
   | FUNCTION NAME: getparm.c                                   |
   | Copyright (C) 2021 P. Michaels  (See GNU License)          |
   |------------------------------------------------------------|
   |                                                            |
   | purpose: gets input parameters from command                |
   |          line or prompts user from consol.                 |
   |                                                            |
   |                                                            |
   |------------------------------------------------------------|
*/
      int  getparm(int argc,char *argv[],
         char infile[],
         char ofile[], char lstfile[], int *n1, int *nlast, char pid[])
     {
      char              sfx1[4]="txt",
                        sfx2[4]="lst";
      char		help[3]="-h";
      char              msg1[46],msg2[46];

//...test for command line help ?
      if(argc==2 && !strcmp(argv[1],help))
     {
//...screen header and copyright notice
      cr_labl(pid);

//...help pannel
      sprintf(msg1,"Audio WAV to ASCII text conversion program   ");
      sprintf(msg2,"                                             ");
      hlp_labl(msg1,msg2);
      puts(" ");

//...help display
      puts(" ");
      printf(" wav2txt  infile n1  nlast   \n");
      puts(" ");
      puts("     infile  = input file name (4char minimum)");
      puts("     n1   = first sample to output  ");
      puts("     nlast    = last  sample to output");
      puts(" ");
      puts(" ");
      return(1);
     }
      else
     {
//...get parameters------------------------------
//...pars off character parameters
      if(argc>1)
       strncpy(infile,argv[1],80);
       else
      {
       puts(" Enter INPUT file name ");
       scanf(" %80s",infile);
      }
	if (argc>2)
	*n1 = atoi(argv[2]);
	else
	{
	puts(" Enter First sample number to extract");
	scanf(" %d",n1);
	}
	if (argc>3)
	*nlast = atoi(argv[3]);
	else
	{
	puts(" Enter Last sample number to extract");
	scanf(" %d",nlast);
	}
//	if (namecheck(12,infile) != 0 ) return (1);

///...name output files
      names(infile,pid,ofile,sfx1);
      names(infile,pid,lstfile,sfx2);
       return(0);
     }  //  endif bhelp test
      } // end getparm
/* |------------------------------------------------------------|
   | Basic Seismic Utilities (BSU)                              | 
   | LANGUAGE: C                                                |
   | AUTHOR: P. MICHAELS <paulmichaels@boisestate.boisestate.edu>|
   | FUNCTION NAME: outlst.c                                    |
   | Copyright (C) 2021 P. Michaels  (See GNU License)          |
   |------------------------------------------------------------|
   |                                                            |
   | purpose: prints copyright header, file names, and          |
   |          echo check of input parameters in newly           |
   |          opened output listing file.                       |
   |                                                            |
   |------------------------------------------------------------|
*/
      void outlst(char infile[],
        char pid[], char ofile[], char lstfile[], FILE *h3)
     {

//...print copyright label to outlst and screen

      fcr_labl(pid,h3);

//...echo check
      fputs("\n",h3);                      //blank line
      fprintf(h3,"\n     PROGRAM: wav2txt.c ");
      fputs("\n",h3);                      //blank line
      fputs("\n     FILES:       ",h3);
      fprintf(h3,"\n     Input   File= %s",infile);
      fprintf(h3,"\n     Output  File= %s",ofile);
      fprintf(h3,"\n     Listing File= %s",lstfile);
      fputs("\n",h3);                      //blank line
      fputs("\n     WAV HEADER CAPTURE: ",h3);
      fputs("\n",h3);                      //blank line
     }  //  end outlst

	long finddata(int *nbytesdata, FILE *h1  )
	{
	int	jt;
	char 		buf2[5]="data",
			buf1[5];
	long	gostart;

	for (jt=0; jt<80; jt++) {
	if(fread(buf1,sizeof(char),5,h1)!=5)
            {
             fputs("ABORT--: Error looking for data string ",stderr);
             return(1);
            }
	if (strncmp(buf1,buf2,4) == 0 ) goto found;
	fseek(h1,(long) -4, SEEK_CUR);
	} //next jt
	fprintf(stderr,"Failed to find string data \n");
	return((long) -1);
found:
	gostart = ftell(h1) - 1;
//	printf("debug: gostart = %ld \n",gostart);
	fseek(h1,(long) gostart,SEEK_SET);
	fread(nbytesdata,sizeof(int),1,h1);
//	printf("debug: nybtesdata=%d \n",*nbytesdata);
	return(gostart);
	} //end finddata
